#ifndef ASYNC_CLIENT_H
#define ASYNC_CLIENT_H
#include <iostream>
#include <stdio.h>
#include <time.h>
#include <boost/asio.hpp>
#include <boost/bind.hpp>
#include <boost/thread.hpp>
#include <string>
#include <unistd.h>
#include "glog/logging.h"

// using namespace std ;
using boost::asio::ip::udp;
using boost::asio::ip::tcp;
using namespace boost::asio;
#define MAX_LENGTH 50000
typedef void (*fundata_t)(const char* addr,int port,const char* data,const int len, void *p);

/**
 * 异步tcp通信客户端
 * */
class Async_Client
{
public:
  // // 无参构造
  // Async_Client();
  // 有参构造
  Async_Client(boost::asio::io_service& io_service,ip::tcp::endpoint endpoint,fundata_t fundata_ , void *p);
  // 析构
  ~Async_Client();
  // 异步写入
  void client_async_write(char *buf,int len);
  // 异步读取
  void client_async_read();
  // 连接状态返回
  bool client_connection_status();
  // 初始化状态返回
  bool client_initialize_status();
  // 初始化
  bool initialize(bool with_reconnection=true);
  // 连接
  void socket_connect();
  // 关闭连接
  void socket_close();

private:
  // 异步读回调
  void handle_read(const boost::system::error_code& error,
                   size_t bytes_transferred);

  // 异步写回调
  void handle_write(const boost::system::error_code& error);

  // 异步连接回调
  void handle_connect(const boost::system::error_code& error);

  io_service &iosev;                      // 异步控制服务
  ip::tcp::socket socket;                 // socket句柄
  ip::tcp::endpoint m_ep;                 // 连接参数

  char        m_origin_data_[MAX_LENGTH];          // 原始数据
  bool        mb_with_reconnection;       // 开启自动重连 
  bool        mb_connected;               // 连接状态
  bool        mb_initialized;             // 初始化是否成功
  fundata_t   m_fundata ;                 // 数据处理回调
  void*       mp_handle;                  // 外部句柄

};

#endif // ASYNC_CLIENT_H
